#include <bits/stdc++.h>
#include "common.h"

struct SubtaskConstraints : BaseConstraints {
	int64_t MAX_N;

	void validate(std::istream& in) const {
		int64_t n;
		in >> n;

		assert(2 <= n && n <= MAX_N);

		std::vector<int64_t> perm(n);
		for (int64_t i = 0; i < n; i ++) {
			in >> perm[i];
		}
		std::sort(perm.begin(), perm.end());
		for (int64_t i = 0; i < n; i ++) {
			assert(perm[i] == i);
		}
	}
};

bool generateSimple(
	std::ostream& out,
	const std::optional<SubtaskConstraints>& constraints
) {
	const auto constr = constraints.value();
	const int64_t n = constr.MAX_N;
	std::vector<int64_t> perm;
	for (int64_t i = 0; i < n; i ++) { perm.push_back(i); }
	shuffle(perm.begin(), perm.end());
	out << n << "\n";
	for (const auto it : perm) { out << it << " "; }
	out << "\n";
	return true;
}

bool generateBigCycle(
	std::ostream& out,
	const std::optional<SubtaskConstraints>& constraints,
	const int64_t leftout
) {
	const auto constr = constraints.value();
	const int64_t n = constr.MAX_N;
	std::vector<int64_t> perm, all;
	for (int64_t i = 0; i < n; i ++) {
		all.push_back(i);
		perm.push_back(i);
	}
	shuffle(all.begin(), all.end());
	for (int64_t i = 0; i < leftout; i ++) {
		all.pop_back();
	}
	for (int64_t i = 0; i < all.size(); i ++) {
		perm[all[(i + 1) % all.size()]] = all[i];
	}
	out << n << "\n";
	for (const auto it : perm) { out << it << " "; }
	out << "\n";
	return true;
}

bool generateBigCyclePref(
	std::ostream& out,
	const std::optional<SubtaskConstraints>& constraints,
	const int64_t leftout,
	const bool prefix
) {
	const auto constr = constraints.value();
	const int64_t n = constr.MAX_N;
	std::vector<int64_t> perm, all;
	for (int64_t i = 0; i < n; i ++) {
		perm.push_back(i);
	}
	if (prefix) {
		for (int64_t i = 0; i < n - leftout; i ++) {
			all.push_back(i);
		}
	} else {
		for (int64_t i = leftout; i < n; i ++) {
			all.push_back(i);
		}
	}
	shuffle(all.begin(), all.end());
	for (int64_t i = 0; i < all.size(); i ++) {
		perm[all[(i + 1) % all.size()]] = all[i];
	}
	out << n << "\n";
	for (const auto it : perm) { out << it << " "; }
	out << "\n";
	return true;
}

bool generatePair(
	std::ostream& out,
	const std::optional<SubtaskConstraints>& constraints,
	const int64_t weight
) {
	const auto constr = constraints.value();
	const int64_t n = constr.MAX_N;
	std::vector<int64_t> perm, all;
	for (int64_t i = 0; i < n; i ++) {
		perm.push_back(i);
	}
	int64_t a = 0, b = 0;
	while (a == b) {
		a = rnd.wnext(n, -weight);
		b = rnd.wnext(n, weight);
	}
	std::swap(perm[a], perm[b]);
	out << n << "\n";
	for (const auto it : perm) { out << it << " "; }
	out << "\n";
	return true;
}

bool generateEvenOdd(
	std::ostream& out,
	const std::optional<SubtaskConstraints>& constraints,
	const int64_t weight
) {
	const auto constr = constraints.value();
	const int64_t n = constr.MAX_N;
	std::vector<int64_t> perm, all;
	for (int64_t i = 0; i < n; i ++) {
		perm.push_back(i);
	}
	for (int64_t i = 0; i + 1 < n; i += 2) {
		if (rnd.wnext(2, weight)) {
			std::swap(perm[i], perm[i + 1]);
		}
	}
	out << n << "\n";
	for (const auto it : perm) { out << it << " "; }
	out << "\n";
	return true;
}

signed main(int argc, char **argv) {
	registerGen(argc, argv, 1);

	const auto subtask1 =
		Subtask<SubtaskConstraints>("MAX_N <= 1000", 20).
		withConstraints({
			.MAX_N = 1000,
		}).
		withTestGens(3, nameFun(generateEvenOdd), std::vector<int64_t>({0, 1, 2, 4, 8, 16}));

	const auto subtask2 =
		Subtask<SubtaskConstraints>("MAX_N <= 1000", 20).
		withConstraints({
			.MAX_N = 1000,
		}).
		withTestGens(3, nameFun(generatePair), std::vector<int64_t>({0, 1, 2, 4, 8, 16}));

	const auto subtask3 =
		Subtask<SubtaskConstraints>("MAX_N <= 1000", 60).
		withConstraints({
			.MAX_N = 1000,
		}).
		withTestGens(10, nameFun(generateSimple)).
		withTestGens(1, nameFun(generateBigCycle), std::vector<int>({0, 1, 2, 4, 8, 16, 32, 64, 128, 256})).
		withTestGens(1, nameFun(generateBigCyclePref), std::vector<int>({0, 1, 2, 4, 8, 16, 32, 64, 128, 256}), std::vector<bool>({false, true}));

	const std::vector<Subtask<SubtaskConstraints>> subtasks = {
		subtask1,
		subtask2,
		subtask3
	};

	generateSubtasks("cycles", subtasks, Solution("author_qsort"));
	generateGradeProperties(subtasks, "256", "1.0");
}
